use image::imageops::overlay;
use image::{DynamicImage, GenericImageView, RgbaImage};

use crate::config::image_processors::ProcessorParams;
use crate::config::Config;
use crate::display::get_max_resolution;
use crate::result::MBGResult;
use rayon::prelude::*;
use std::collections::hash_map::RandomState;
use std::collections::HashMap;

pub mod blur;
pub mod border;
pub mod crop;
pub mod none;
pub mod scale;

#[derive(Serialize, Clone, Deserialize, Debug)]
pub enum Processors {
    None,
    Blur,
    Scale,
    Crop {
        #[serde(default)]
        width: u32,
        #[serde(default)]
        height: u32,
        #[serde(default)]
        fit: bool,
    },
    Border {
        #[serde(default)]
        circle: bool,
        #[serde(default)]
        color: String,
        #[serde(default)]
        width: u32,
    },
}

pub trait ImageProcessor {
    fn process(&self, img: &DynamicImage) -> MBGResult<DynamicImage>;
}

pub trait GetLayer {
    fn get_layer(&self) -> usize;
}

pub trait ImgFilterTrait: GetLayer + ImageProcessor {}

macro_rules! impl_params_getter {
    ($($inst:path),*) => {
        $(impl GetLayer for $inst {
            fn get_layer(&self) -> usize {
                self.params.layer
            }
        }
        impl ImgFilterTrait for $inst{}
        )*
    };
}

impl Default for Processors {
    fn default() -> Self {
        Self::None
    }
}

pub fn process_layer(
    screen_height: i32,
    screen_width: i32,
    image: DynamicImage,
    processors: Vec<ProcessorParams>,
) -> MBGResult<DynamicImage> {
    let mut image = image;
    for processor_params in processors {
        let processor = mapper(processor_params, screen_height, screen_width);
        image = processor.process(&image)?;
        debug!("Processor applied")
    }
    Ok(image)
}

pub fn process_image(img_bytes: Vec<u8>) -> MBGResult<DynamicImage> {
    let conf = Config::get_props();
    let (screen_height, screen_width) = get_max_resolution()?;
    debug!("Maximum resolution: {}x{}", screen_width, screen_height);
    let image = image::load_from_memory(img_bytes.as_slice())?;
    let layers_conf = conf.processors.iter().fold(
        HashMap::<usize, Vec<ProcessorParams>>::new(),
        |mut acc, x| {
            let new_processors = if let Some(current) = acc.get(&x.layer) {
                let mut mutable = current.clone();
                mutable.push(x.clone());
                mutable
            } else {
                vec![x.clone()]
            };
            acc.insert(x.layer, new_processors);
            acc
        },
    );
    let processed_layers = layers_conf
        .par_iter()
        .map(|(layer, processors)| {
            let img = process_layer(
                screen_height,
                screen_width,
                image.clone(),
                processors.to_vec(),
            );
            (layer, img)
        })
        .collect::<HashMap<_, _, RandomState>>();
    blend_layers(screen_height, screen_width, processed_layers, conf.blender)
}

pub fn blend_layers<S: std::hash::BuildHasher>(
    height: i32,
    width: i32,
    layers: HashMap<&usize, MBGResult<DynamicImage>, S>,
    blender: Vec<u8>,
) -> MBGResult<DynamicImage> {
    let mut res = DynamicImage::ImageRgba8(RgbaImage::new(width as u32, height as u32));
    for index in blender.into_iter() {
        if let Some(layer) = layers.get(&usize::from(index)) {
            match layer {
                Ok(image) => {
                    let pos_x = (res.width() / 2) - (image.width() / 2);
                    let pos_y = (res.height() / 2) - (image.height() / 2);
                    overlay(&mut res, image, pos_x, pos_y);
                }
                Err(error) => {
                    error!("{}", error);
                }
            }
        }
    }
    Ok(res)
}

impl_params_getter!(
    none::NoneProcessor,
    blur::BlurProcessor,
    scale::ScaleProcessor,
    crop::CropProcessor,
    border::Border
);

fn mapper(
    params: ProcessorParams,
    screen_height: i32,
    screen_width: i32,
) -> Box<dyn ImgFilterTrait> {
    match params.clone().effect {
        Processors::None => Box::new(none::NoneProcessor { params }),
        Processors::Blur => Box::new(blur::BlurProcessor { params }),
        Processors::Scale => Box::new(scale::ScaleProcessor {
            params,
            screen_height,
            screen_width,
        }),
        Processors::Crop { width, height, fit } => Box::new(crop::CropProcessor {
            params,
            screen_height,
            screen_width,
            height,
            width,
            fit,
        }),
        Processors::Border {
            circle,
            color,
            width,
        } => Box::new(border::Border {
            params,
            circle,
            color,
            width,
        }),
    }
}
